/************************************************************************/
/* gtc_p.h - GPL Telescope Control Library Internal Header		*/
/************************************************************************/
/*
 *  (c) 2004 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/

#ifndef _GTC_P_H_
#define _GTC_P_H_


#include <gtc/gtctypes.h>


/* GTC error code range */
#define GTC_RC_BASEERROR		1000
#define GTC_RC_MAXERROR			1010


/* GTC device flags (telescope and observatory) */
#define	GTC_FLAG_DEVICECONNECTED	0x00000001	/* a device is connected */


typedef struct	_GtcModuleInfo		GtcModuleInfo;
typedef struct	_GtcVector		GtcVector;


typedef int (*GtcLoadFunction) (GtcModuleInfo *gmi);
typedef int (*GtcUnloadFunction) (GtcModuleInfo *gmi);


struct _GtcModuleInfo {
	unsigned long	reserved;	/* reserved for future use */
	GtcVector	*vector;	/* module vectors */
	char		*name;		/* module's short name (no whitespace) */
	char		*description;	/* module's description */
	char		*version;	/* module's version */
	char		*author;	/* module's author */
	void		*data;		/* module specific data */
	unsigned long	capabilities;	/* module capability flags */
};


struct _GtcModule {
	unsigned long	reserved;	/* reserved for future use */
	void		*module;	/* dlopen() module */
	GtcUnloadFunction unload;	/* module unload function */
	GtcModuleInfo	*moduleinfo;	/* module information */
	char		*path;		/* module path (dynamic allocation) */
	int		refcount;	/* module reference counter */
};


struct _GtcScope {
	unsigned long	reserved;	/* reserved for future use */
	GtcModule	*module;	/* module this scope uses */
	char		*device;	/* device path for the scope */
	void		*data;		/* telescope specific data */
	unsigned long	flags;		/* device specific flags */
};


struct _GtcObservatory {
	unsigned long	reserved;	/* reserved for future use */
	GtcModule	*module;	/* module this observatory uses */
	char		*device;	/* device path for the observatory */
	void		*data;		/* observatory specific data */
	unsigned long	flags;		/* device specific flags */
};


struct _GtcVector {
	unsigned long	reserved;	/* reserved for future use */

	/* scope connect/disconnect */
	int (*scope_connect) (GtcScope *gs, char *device);
	int (*scope_close) (GtcScope *gs);

	/* scope model/firmware information */
	int (*scope_get_model) (GtcScope *gs, char **model);
	int (*scope_get_firmware_version) (GtcScope *gs, char **version);
	int (*scope_get_firmware_date) (GtcScope *gs, char **date);

	/* scope date/time */
	int (*scope_get_datetime) (GtcScope *gs, GtcDate *gd, GtcTime *gt);
	int (*scope_set_datetime) (GtcScope *gs, GtcDate *gd, GtcTime *gt);
	int (*scope_get_utc_offset) (GtcScope *gs, double *offset);
	int (*scope_set_utc_offset) (GtcScope *gs, double offset);
	int (*scope_get_sidereal_time) (GtcScope *gs, GtcTime *gt);
	int (*scope_set_sidereal_time) (GtcScope *gs, GtcTime *gt);

	/* scope mount/site/location */
	int (*scope_autoalign) (GtcScope *gs);
	int (*scope_home) (GtcScope *gs);
	int (*scope_get_alignment) (GtcScope *gs, int *alignment);
	int (*scope_set_alignment) (GtcScope *gs, int alignment);
	int (*scope_get_longitude) (GtcScope *gs, double *longitude);
	int (*scope_set_longitude) (GtcScope *gs, double longitude);
	int (*scope_get_latitude) (GtcScope *gs, double *latitude);
	int (*scope_set_latitude) (GtcScope *gs, double latitude);
	int (*scope_get_current_site) (GtcScope *gs, int *site);
	int (*scope_set_current_site) (GtcScope *gs, int site);

	/* scope RA/DEC motion */
	int (*scope_slew_east_start) (GtcScope *gs);
	int (*scope_slew_east_stop) (GtcScope *gs);
	int (*scope_slew_west_start) (GtcScope *gs);
	int (*scope_slew_west_stop) (GtcScope *gs);
	int (*scope_slew_north_start) (GtcScope *gs);
	int (*scope_slew_north_stop) (GtcScope *gs);
	int (*scope_slew_south_start) (GtcScope *gs);
	int (*scope_slew_south_stop) (GtcScope *gs);
	int (*scope_slew_stop) (GtcScope *gs);

	int (*scope_get_ra_slewrate) (GtcScope *gs, double *degsec);
	int (*scope_set_ra_slewrate) (GtcScope *gs, double degsec); 
	int (*scope_get_dec_slewrate) (GtcScope *gs, double *degsec);
	int (*scope_set_dec_slewrate) (GtcScope *gs, double degsec); 

	int (*scope_get_trackrate) (GtcScope *gs, double *hertz);
	int (*scope_set_trackrate) (GtcScope *gs, double hertz);

	int (*scope_get_guiderate) (GtcScope *gs, double *arcsecsec);
	int (*scope_set_guiderate) (GtcScope *gs, double arcsecsec);

	/* scope focus */
	int (*scope_focus_in_start) (GtcScope *gs);
	int (*scope_focus_out_start) (GtcScope *gs);
	int (*scope_focus_stop) (GtcScope *gs);
	int (*scope_focus_get_speed) (GtcScope *gs, double *speed);
	int (*scope_focus_set_speed) (GtcScope *gs, double speed);

	/* scope reticle */
	int (*scope_reticle_get_brightness) (GtcScope *gs, double *brightness);
	int (*scope_reticle_set_brightness) (GtcScope *gs, double brightness);
	int (*scope_reticle_get_flashrate) (GtcScope *gs, double *rate);
	int (*scope_reticle_set_flashrate) (GtcScope *gs, double rate);
	int (*scope_reticle_get_dutyrate) (GtcScope *gs, double *rate);
	int (*scope_reticle_set_dutyrate) (GtcScope *gs, double rate);

	/* scope field derotator */
	int (*scope_derotator_enable) (GtcScope *gs, int enable);

	/* scope targeting/tracking */
	int (*scope_get_current_ra) (GtcScope *gs, GtcRa *ra);
	int (*scope_get_current_dec) (GtcScope *gs, GtcDec *dec);

	int (*scope_get_target_ra) (GtcScope *gs, GtcRa *ra);
	int (*scope_set_target_ra) (GtcScope *gs, GtcRa *ra);
	int (*scope_get_target_dec) (GtcScope *gs, GtcDec *dec);
	int (*scope_set_target_dec) (GtcScope *gs, GtcDec *dec);
	int (*scope_goto_target) (GtcScope *gs);

	/* scope library objects */
	int (*scope_object_set_target) (GtcScope *gs, int catalog, long object);

	/* scope Periodic Error Correction (PEC) */
	int (*scope_pec_ra_enable) (GtcScope *gs, int enable);
	int (*scope_pec_dec_enable) (GtcScope *gs, int enable);

	/* scope options */
	int (*scope_option_list) (GtcScope *gs, GtcOption **optionlist);
	int (*scope_option_get_int) (GtcScope *gs, char *name, int *value);
	int (*scope_option_set_int) (GtcScope *gs, char *name, int value);
	int (*scope_option_get_boolean) (GtcScope *gs, char *name, int *value);
	int (*scope_option_set_boolean) (GtcScope *gs, char *name, int value);
	int (*scope_option_get_double) (GtcScope *gs, char *name, double *value);
	int (*scope_option_set_double) (GtcScope *gs, char *name, double value);
	int (*scope_option_get_string) (GtcScope *gs, char *name, char **value);
	int (*scope_option_set_string) (GtcScope *gs, char *name, char *value);

	/* observatory control elements */
	int (*observatory_connect) (GtcObservatory *go, char *device);
	int (*observatory_close) (GtcObservatory *go);
	int (*observatory_slew_cw_start) (GtcObservatory *go);
	int (*observatory_slew_ccw_start) (GtcObservatory *go);
	int (*observatory_slew_stop) (GtcObservatory *go);
	int (*observatory_skylight_open) (GtcObservatory *go);
	int (*observatory_skylight_close) (GtcObservatory *go);
	int (*observatory_light_enable) (GtcObservatory *go, int enable);
	int (*observatory_nightlight_enable) (GtcObservatory *go, int enable);

	/* Retreive telescope control values */
	int (*scope_control_get_slewrates) (GtcScope *gs, GtcControl **gcp);
	int (*scope_control_get_trackrates) (GtcScope *gs, GtcControl **gcp);
	int (*scope_control_get_guiderates) (GtcScope *gs, GtcControl **gcp);
	int (*scope_control_get_focus_speeds) (GtcScope *gs, GtcControl **gcp);
	int (*scope_control_get_reticle_flashrates) (GtcScope *gs, GtcControl **gcp);
	int (*scope_control_get_reticle_dutyrates) (GtcScope *gs, GtcControl **gcp);

};



#endif

