/************************************************************************/
/* gtc_meade.c - GPL Telescope Control Library Plugin Source (Meade)	*/
/************************************************************************/
/*
 *  (c) 2004 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/select.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <termios.h>
#include <math.h>

#include <gtc/gtctypes.h>
#include <gtc/gtc_p.h>
#include <gtc/gtcdef.h>


#define	GTC_NAME		"meade"
#define	GTC_DESCRIPTION		"Meade Telescope Protocol"
#define	GTC_VERSION		"0.0.2"
#define	GTC_AUTHOR		"Todd Shadburn"

#define BUFFER_SIZE		128

#define CMDFLAG_NORESPONSE	0x01 /* don't expect command to return data */
#define CMDFLAG_WAIT5MINUTES	0x02 /* make the response timeout 5 minutes */
#define CMDFLAG_NOTIMEOUT	0x04 /* wait FOREVER for the response data */
#define CMDFLAG_NOENDHASH	0x04 /* the data will not end with a '#' */

#define MEADE_MODEL_LX200GPS	0x01
#define MEADE_MODEL_AUTOSTAR	0x02
#define MEADE_MODEL_LX200	0x03
#define MEADE_MODEL_LX16	0x04


typedef struct _MeadeScope	MeadeScope;
struct _MeadeScope {
	int	fd;			/* serial connection file descriptor */
	char	buffer[BUFFER_SIZE];	/* command buffer */
	struct	termios t_orig;		/* original terminal(serial) settings */
	int	model;			/* Meade telescope model */
};


GtcOption meade_options_lx200gps[] = {
	{"Gps", GTC_OPTION_TYPE_BOOLEAN, 0, 1, 0, 0},
	{"AuxPower", GTC_OPTION_TYPE_BOOLEAN, 0, 1, 0, 0},
	{"OTATemp(C)", GTC_OPTION_TYPE_DOUBLE, -127.0, 500.0, 0, GTC_OPTION_FLAG_READONLY},
};

GtcControlValue meade_slewrate_values[] = {
	{"1x", 1},
	{"2x", 2},
	{"8x", 3},
	{"16x", 4},
	{"64x", 5},
	{"128x", 6},
	{"1.5 deg/sec", 7},
	{"3 deg/sec", 8},
	{"8 deg/sec", 9},
};
GtcControl meade_slewrates = {0, 1.0, 9.0, 9.0, meade_slewrate_values};

GtcControlValue meade_trackrate_values[] = {
	{"Sidereal", 60.0},
	{"Solar", 60.0},
	{"Lunar", 60.0},
};
GtcControl meade_trackrates = {0, 0.1, 999.9, 60.0, meade_trackrate_values};

GtcControl meade_guiderates = {0, -15.0, 15.0, 0.0, NULL};

GtcControlValue meade_focus_speed_values[] = {
	{"Fine", 1},
	{"Slow", 2},
	{"Medium", 3},
	{"Fast", 4},
};
GtcControl meade_focus_speeds = {0, 1, 4, 4, meade_focus_speed_values};

GtcControl meade_reticle_flashrates_lx200 = {0, 0, 3, 2, NULL};
GtcControl meade_reticle_flashrates_lx200gps= {0, 0, 9, 4, NULL};

GtcControl meade_reticle_dutyrates = {0, 0, 15, 0, NULL};

	

/* Callback function prototypes */
int degrees_decimal_to_dms (double decdegree, int *degree,
				int *arcminute, int *arcsecond);
int degrees_dms_to_decimal (double *decdegree, int degree,
				int arcminute, int arcsecond);
int command_execute (MeadeScope *ms, int cmdflag);
int meade_scope_connect (GtcScope *gs, char *device);
int meade_scope_close (GtcScope *gs);
int meade_scope_get_model (GtcScope *gs, char **model);
int meade_scope_get_firmware_version (GtcScope *gs, char **version);
int meade_scope_get_firmware_date (GtcScope *gs, char **date);
int meade_scope_get_datetime (GtcScope *gs, GtcDate *gd, GtcTime *gt);
int meade_scope_set_datetime (GtcScope *gs, GtcDate *gd, GtcTime *gt);
int meade_scope_get_utc_offset (GtcScope *gs, double *offset);
int meade_scope_set_utc_offset (GtcScope *gs, double offset);
int meade_scope_get_sidereal_time (GtcScope *gs, GtcTime *gt);
int meade_scope_set_sidereal_time (GtcScope *gs, GtcTime *gt);
int meade_scope_autoalign (GtcScope *gs);
int meade_scope_home (GtcScope *gs);
int meade_scope_get_alignment (GtcScope *gs, int *alignment);
int meade_scope_set_alignment (GtcScope *gs, int alignment);
int meade_scope_get_longitude (GtcScope *gs, double *longitude);
int meade_scope_set_longitude (GtcScope *gs, double longitude);
int meade_scope_get_latitude (GtcScope *gs, double *latitude);
int meade_scope_set_latitude (GtcScope *gs, double latitude);
int meade_scope_get_current_site (GtcScope *gs, int *site);
int meade_scope_set_current_site (GtcScope *gs, int site);
int meade_scope_slew_east_start (GtcScope *gs);
int meade_scope_slew_east_stop (GtcScope *gs);
int meade_scope_slew_west_start (GtcScope *gs);
int meade_scope_slew_west_stop (GtcScope *gs);
int meade_scope_slew_north_start (GtcScope *gs);
int meade_scope_slew_north_stop (GtcScope *gs);
int meade_scope_slew_south_start (GtcScope *gs);
int meade_scope_slew_south_stop (GtcScope *gs);
int meade_scope_slew_stop (GtcScope *gs);
int meade_scope_get_ra_slewrate (GtcScope *gs, double *degsec);
int meade_scope_set_ra_slewrate (GtcScope *gs, double degsec); 
int meade_scope_get_dec_slewrate (GtcScope *gs, double *degsec);
int meade_scope_set_dec_slewrate (GtcScope *gs, double degsec); 
int meade_scope_get_trackrate (GtcScope *gs, double *hertz);
int meade_scope_set_trackrate (GtcScope *gs, double hertz);
int meade_scope_get_guiderate (GtcScope *gs, double *arcsecsec);
int meade_scope_set_guiderate (GtcScope *gs, double arcsecsec);
int meade_scope_focus_in_start (GtcScope *gs);
int meade_scope_focus_out_start (GtcScope *gs);
int meade_scope_focus_stop (GtcScope *gs);
int meade_scope_focus_get_speed (GtcScope *gs, double *speed);
int meade_scope_focus_set_speed (GtcScope *gs, double speed);
int meade_scope_reticle_get_brightness (GtcScope *gs, double *brightness);
int meade_scope_reticle_set_brightness (GtcScope *gs, double brightness);
int meade_scope_reticle_get_flashrate (GtcScope *gs, double *rate);
int meade_scope_reticle_set_flashrate (GtcScope *gs, double rate);
int meade_scope_reticle_get_dutyrate (GtcScope *gs, double *rate);
int meade_scope_reticle_set_dutyrate (GtcScope *gs, double rate);
int meade_scope_derotator_enable (GtcScope *gs, int enable);
int meade_scope_get_current_ra (GtcScope *gs, GtcRa *ra);
int meade_scope_get_current_dec (GtcScope *gs, GtcDec *dec);
int meade_scope_get_target_ra (GtcScope *gs, GtcRa *ra);
int meade_scope_set_target_ra (GtcScope *gs, GtcRa *ra);
int meade_scope_get_target_dec (GtcScope *gs, GtcDec *dec);
int meade_scope_set_target_dec (GtcScope *gs, GtcDec *dec);
int meade_scope_goto_target (GtcScope *gs);
int meade_scope_object_set_target (GtcScope *gs, int catalog, long object);
int meade_scope_pec_ra_enable (GtcScope *gs, int enable);
int meade_scope_pec_dec_enable (GtcScope *gs, int enable);
int meade_scope_option_list (GtcScope *gs, GtcOption **optionlist);
int meade_scope_option_get_int (GtcScope *gs, char *name, int *value);
int meade_scope_option_set_int (GtcScope *gs, char *name, int value);
int meade_scope_option_get_double (GtcScope *gs, char *name, double *value);
int meade_scope_option_set_double (GtcScope *gs, char *name, double value);
int meade_scope_option_get_string (GtcScope *gs, char *name, char **value);
int meade_scope_option_set_string (GtcScope *gs, char *name, char *value);

int meade_scope_control_get_slewrates (GtcScope *gs, GtcControl **gcp);
int meade_scope_control_get_trackrates (GtcScope *gs, GtcControl **gcp);
int meade_scope_control_get_guiderates (GtcScope *gs, GtcControl **gcp);
int meade_scope_control_get_focus_speeds (GtcScope *gs, GtcControl **gcp);
int meade_scope_control_get_reticle_flashrates (GtcScope *gs, GtcControl **gcp);
int meade_scope_control_get_reticle_dutyrates (GtcScope *gs, GtcControl **gcp);



int module_load (GtcModuleInfo *gmi)
{
	/* NULL all callbacks */
	memset (gmi->vector, 0, sizeof (GtcVector));

	/* set the implemented callbacks */
	gmi->vector->scope_connect = meade_scope_connect;
	gmi->vector->scope_close = meade_scope_close;
	gmi->vector->scope_get_model = meade_scope_get_model;
	gmi->vector->scope_get_firmware_version = meade_scope_get_firmware_version;
	gmi->vector->scope_get_firmware_date = meade_scope_get_firmware_date;
	gmi->vector->scope_get_datetime = meade_scope_get_datetime;
	gmi->vector->scope_set_datetime = meade_scope_set_datetime;
	gmi->vector->scope_get_utc_offset = meade_scope_get_utc_offset;
	gmi->vector->scope_set_utc_offset = meade_scope_set_utc_offset;
	gmi->vector->scope_get_sidereal_time = meade_scope_get_sidereal_time;
	gmi->vector->scope_set_sidereal_time = meade_scope_set_sidereal_time;
	gmi->vector->scope_autoalign = meade_scope_autoalign;
	gmi->vector->scope_home = meade_scope_home;
	gmi->vector->scope_get_alignment = meade_scope_get_alignment;
	gmi->vector->scope_set_alignment = meade_scope_set_alignment;
	gmi->vector->scope_get_longitude = meade_scope_get_longitude;
	gmi->vector->scope_set_longitude = meade_scope_set_longitude;
	gmi->vector->scope_get_latitude = meade_scope_get_latitude;
	gmi->vector->scope_set_latitude = meade_scope_set_latitude;
	gmi->vector->scope_get_current_site = meade_scope_get_current_site;
	gmi->vector->scope_set_current_site = meade_scope_set_current_site;
	gmi->vector->scope_slew_east_start = meade_scope_slew_east_start;
	gmi->vector->scope_slew_east_stop = meade_scope_slew_east_stop;
	gmi->vector->scope_slew_west_start = meade_scope_slew_west_start;
	gmi->vector->scope_slew_west_stop = meade_scope_slew_west_stop;
	gmi->vector->scope_slew_north_start = meade_scope_slew_north_start;
	gmi->vector->scope_slew_north_stop = meade_scope_slew_north_stop;
	gmi->vector->scope_slew_south_start = meade_scope_slew_south_start;
	gmi->vector->scope_slew_south_stop = meade_scope_slew_south_stop;
	gmi->vector->scope_slew_stop = meade_scope_slew_stop;
	gmi->vector->scope_get_ra_slewrate = meade_scope_get_ra_slewrate;
	gmi->vector->scope_set_ra_slewrate = meade_scope_set_ra_slewrate;
	gmi->vector->scope_get_dec_slewrate = meade_scope_get_dec_slewrate;
	gmi->vector->scope_set_dec_slewrate = meade_scope_set_dec_slewrate;
	gmi->vector->scope_get_trackrate = meade_scope_get_trackrate;
	gmi->vector->scope_set_trackrate = meade_scope_set_trackrate;
	gmi->vector->scope_get_guiderate = meade_scope_get_guiderate;
	gmi->vector->scope_set_guiderate = meade_scope_set_guiderate;
	gmi->vector->scope_focus_in_start = meade_scope_focus_in_start;
	gmi->vector->scope_focus_out_start = meade_scope_focus_out_start;
	gmi->vector->scope_focus_stop = meade_scope_focus_stop;
	gmi->vector->scope_focus_get_speed = meade_scope_focus_get_speed;
	gmi->vector->scope_focus_set_speed = meade_scope_focus_set_speed;
	gmi->vector->scope_reticle_get_brightness = meade_scope_reticle_get_brightness;
	gmi->vector->scope_reticle_set_brightness = meade_scope_reticle_set_brightness;
	gmi->vector->scope_reticle_get_flashrate = meade_scope_reticle_get_flashrate;
	gmi->vector->scope_reticle_set_flashrate = meade_scope_reticle_set_flashrate;
	gmi->vector->scope_reticle_get_dutyrate = meade_scope_reticle_get_dutyrate;
	gmi->vector->scope_reticle_set_dutyrate = meade_scope_reticle_set_dutyrate;
	gmi->vector->scope_derotator_enable = meade_scope_derotator_enable;
	gmi->vector->scope_get_current_ra = meade_scope_get_current_ra;
	gmi->vector->scope_get_current_dec = meade_scope_get_current_dec;
	gmi->vector->scope_get_target_ra = meade_scope_get_target_ra;
	gmi->vector->scope_set_target_ra = meade_scope_set_target_ra;
	gmi->vector->scope_get_target_dec = meade_scope_get_target_dec;
	gmi->vector->scope_set_target_dec = meade_scope_set_target_dec;
	gmi->vector->scope_goto_target = meade_scope_goto_target;
	gmi->vector->scope_object_set_target = meade_scope_object_set_target;
	gmi->vector->scope_pec_ra_enable = meade_scope_pec_ra_enable;
	gmi->vector->scope_pec_dec_enable = meade_scope_pec_dec_enable;
	gmi->vector->scope_option_list = meade_scope_option_list;
	gmi->vector->scope_option_get_int = meade_scope_option_get_int;
	gmi->vector->scope_option_set_int = meade_scope_option_set_int;
	gmi->vector->scope_option_get_double = meade_scope_option_get_double;
	gmi->vector->scope_option_set_double = meade_scope_option_set_double;
	gmi->vector->scope_option_get_string = meade_scope_option_get_string;
	gmi->vector->scope_option_set_string = meade_scope_option_set_string;
	gmi->vector->scope_control_get_slewrates = meade_scope_control_get_slewrates;
	gmi->vector->scope_control_get_trackrates = meade_scope_control_get_trackrates;
	gmi->vector->scope_control_get_guiderates = meade_scope_control_get_guiderates;
	gmi->vector->scope_control_get_focus_speeds = meade_scope_control_get_focus_speeds;
	gmi->vector->scope_control_get_reticle_flashrates = meade_scope_control_get_reticle_flashrates;
	gmi->vector->scope_control_get_reticle_dutyrates = meade_scope_control_get_reticle_dutyrates;


	/* set the module information */
	gmi->name = GTC_NAME;
	gmi->description = GTC_DESCRIPTION;
	gmi->version = GTC_VERSION;
	gmi->author = GTC_AUTHOR;
	return 0;
}


int module_unload (GtcModuleInfo *gmi)
{
	/* NULL all callbacks */
	memset (gmi->vector, 0, sizeof (GtcVector));
	gmi->name = NULL;
	gmi->version = NULL;
	gmi->author = NULL;
	return 0;
}


int degrees_decimal_to_dms (double decdegree, int *degree,
				int *arcminute, int *arcsecond)
{
	double d;
	if (degree == NULL || arcminute == NULL || arcsecond == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*degree = floor (decdegree);
	d = (decdegree - (double)*degree) * 60.0;
	*arcminute = floor (d);
	*arcsecond = (d - (double)*arcminute) * 60.0;
	return 0;
}


int degrees_dms_to_decimal (double *decdegree, int degree,
				int arcminute, int arcsecond)
{
	double d;
	if (decdegree == NULL)
		return GTC_RC_INVALIDARGUMENT;
	d = (double)degree;
	d += (double)arcminute / 60.0;
	d += (double)arcsecond / 3600.0;
	*decdegree = d;
	return 0;
}



int command_execute (MeadeScope *ms, int cmdflag)
{
	int rc;
	int bytes;
	int bytesleft;
	fd_set set;
	struct timeval tv;
	char *p = NULL;
	bytes = strlen (ms->buffer);
	if (bytes >= BUFFER_SIZE)
		return GTC_RC_INVALIDARGUMENT;
	/* write the command data */
fprintf (stderr, "%s: write data '%s'\n", __PRETTY_FUNCTION__, ms->buffer);
	rc = write (ms->fd, ms->buffer, bytes);
	if (rc < bytes)
		return errno;
	if (cmdflag & CMDFLAG_NORESPONSE)
		return 0;
	if ( (cmdflag & CMDFLAG_NOTIMEOUT) != CMDFLAG_NOTIMEOUT) {
		/* use select to wait for I/O on the read fd, or timeout */
		FD_ZERO (&set);
		FD_SET (ms->fd, &set);
		tv.tv_sec = 5;
		if (cmdflag & CMDFLAG_WAIT5MINUTES)
			tv.tv_sec = 300;
		tv.tv_usec = 0;
		rc = select (ms->fd + 1, &set, NULL, NULL, &tv);
		FD_ZERO (&set);
		if (rc == -1)
			return errno;
		if (rc == 0)
			return GTC_RC_TIMEDOUT;
	}
	/* Data arrived, read it */	
	memset (ms->buffer, 0, BUFFER_SIZE);
	bytes = 0;
	p = ms->buffer;
	bytesleft = BUFFER_SIZE - 1;
	do {
		rc = read (ms->fd, p, bytesleft);
		if (rc == -1)
			return errno;
		bytes += rc;
		bytesleft -= rc;
		p += rc;
		if (cmdflag & CMDFLAG_NOENDHASH)
			break; /* the returned data doesn't end with a # sign. */
	} while (bytesleft > 0 && ms->buffer[bytes-1] != '#');
	if (bytesleft == 0)
		return ENOMEM;
	/* remove the trailing '#', if it's there */
	ms->buffer[bytes] = 0;
	if (rc > 0 && ms->buffer[bytes - 1] == '#')
		ms->buffer[bytes - 1] = 0;
fprintf (stderr, "%s: read data(processed) '%s'\n\n", __PRETTY_FUNCTION__, ms->buffer);
	return 0;
}


int meade_scope_connect (GtcScope *gs, char *device)
{
	int rc;
	struct termios t_new;
	MeadeScope *ms = NULL;
	ms = calloc (1, sizeof (MeadeScope));
	if (ms == NULL)
		return GTC_RC_NOMEMORY;
	ms->fd = open (device, O_RDWR);
	if (ms->fd == -1) {
		rc = errno;
		free (ms);
		return rc;
	}

	rc = tcgetattr (ms->fd, &ms->t_orig);
	memcpy (&t_new, &ms->t_orig, sizeof (struct termios));
	t_new.c_iflag = IGNPAR | IGNCR;
	t_new.c_oflag = ONOCR | ONLRET;
	t_new.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
	t_new.c_lflag = 0;
	rc = tcsetattr (ms->fd, TCSANOW, &t_new);
	rc = tcflush (ms->fd, TCIOFLUSH);
	/* Get the product name and test the connection too. */
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GVP#");
	rc = command_execute (ms, 0);
	if (rc != 0) {
		/* This command should have worked, but didn't.
		 * Maybe the scope is not connected?
		*/
		tcsetattr (ms->fd, TCSANOW, &ms->t_orig);
		tcflush (ms->fd, TCIOFLUSH);
		close (ms->fd);
		free (ms);
		return rc;
	}
	if (!strcmp (ms->buffer, "LX2001")) {
		ms->model = MEADE_MODEL_LX200GPS;
	}

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GVN#");
	rc = command_execute (ms, 0);

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GVD#");
	rc = command_execute (ms, 0);

	gs->data = ms;
	return 0;
}


int meade_scope_close (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (ms->fd >= 0) {
		rc = tcsetattr (ms->fd, TCSANOW, &ms->t_orig);
		rc = tcflush (ms->fd, TCIOFLUSH);
		rc = close (ms->fd);
		ms->fd = -1;
	}
	free (ms);
	gs->data = NULL;
	return 0;
}


int meade_scope_get_model (GtcScope *gs, char **model)
{
	int rc;
	MeadeScope *ms = NULL;
	*model = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GVP#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	*model = ms->buffer;
	return 0;
}


int meade_scope_get_firmware_version (GtcScope *gs, char **version)
{
	int rc;
	MeadeScope *ms = NULL;
	*version = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GVN#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	*version = ms->buffer;
	return 0;
}


int meade_scope_get_firmware_date (GtcScope *gs, char **date)
{
	int rc;
	MeadeScope *ms = NULL;
	*date = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GVD#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	*date = ms->buffer;
	return 0;
}


int meade_scope_get_datetime (GtcScope *gs, GtcDate *gd, GtcTime *gt)
{
	int rc;	
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;

	if (gd != NULL) {
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":GC#");
		rc = command_execute (ms, 0);
		if (rc != 0)
			return rc;
		p = ms->buffer;
		d = p;
		d += 2;
		*d = 0;
		gd->month = atoi (p);
		p += 3;
		d += 3;
		*d = 0;
		gd->day = atoi (p);
		p += 3;
		d += 3;
		*d = 0;
		gd->year = atoi (p);
		if (gd->year < 70)
			gd->year += 2000;
		else
			gd->year += 1900;
	}
	if (gt != NULL) {
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":GL#");
		rc = command_execute (ms, 0);
		if (rc != 0)
			return rc;
		p = ms->buffer;
		d = p;
		d += 2;
		*d = 0;
		gt->hour = atoi (p);
		p += 3;
		d += 3;
		*d = 0;
		gt->minute = atoi (p);
		p += 3;
		d += 3;
		*d = 0;
		gt->second = atoi (p);
	}
	return 0;
}


int meade_scope_set_datetime (GtcScope *gs, GtcDate *gd, GtcTime *gt)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (gd != NULL) {
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":SC%02d/%02d/%02d#",
				gd->month, gd->day, gd->year);
		rc = command_execute (ms, 0);
		if (rc != 0)
			return rc;
		if (ms->buffer[0] == '0') /* bad date */
			return GTC_RC_INVALIDARGUMENT;
	}
	if (gt != NULL) {
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":SL%02d:%02d:%02d#",
				gt->hour, gt->minute, gt->second);
		rc = command_execute (ms, CMDFLAG_NOENDHASH);
		if (rc != 0)
			return rc;
		if (ms->buffer[0] == 0) /* bad time */
			return GTC_RC_INVALIDARGUMENT;
	}
	return 0;
}


int meade_scope_get_utc_offset (GtcScope *gs, double *offset)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GG#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	*offset = atof (ms->buffer);
	return 0;
}


int meade_scope_set_utc_offset (GtcScope *gs, double offset)
{
	int rc;	
	char c;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (offset > 0.0)
		c = '+';
	else
		c = '-';
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":SG%c%02d#", c, abs ((int)offset));
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad offset */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_get_sidereal_time (GtcScope *gs, GtcTime *gt)
{
	int rc;	
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GS#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 2;
	*d = 0;
	gt->hour = atoi (p);
	p += 3;
	d += 3;
	*d = 0;
	gt->minute = atoi (p);
	p += 3;
	d += 3;
	*d = 0;
	gt->second = atoi (p);
	return 0;
}


int meade_scope_set_sidereal_time (GtcScope *gs, GtcTime *gt)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":SS%02d:%02d:%02d#",
			gt->hour, gt->minute, gt->second);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad time */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_autoalign (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Aa#");
	rc = command_execute (ms, CMDFLAG_NOTIMEOUT | CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* Not AltAz mounted or align failed */
		return GTC_RC_ALIGNFAILED;
	return 0;
}


int meade_scope_home (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":hP#"); /* Park the scope */
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}



int meade_scope_get_alignment (GtcScope *gs, int *alignment)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	*alignment = 0;
	rc = snprintf (ms->buffer, BUFFER_SIZE, "%c", 0x06);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	switch (ms->buffer[0]) {
		case 'A':
			*alignment = GTC_ALIGNMENT_ALTAZ;
			break;
		case 'P':
			*alignment = GTC_ALIGNMENT_POLAR;
			break;
		case 'L':
			*alignment = GTC_ALIGNMENT_LAND;
			break;
	}
	return 0;
}


int meade_scope_set_alignment (GtcScope *gs, int alignment)
{
	int rc;	
	char a;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (alignment) {
		case GTC_ALIGNMENT_ALTAZ:
			a = 'A';
			break;
		case GTC_ALIGNMENT_POLAR:
			a = 'P';
			break;
		case GTC_ALIGNMENT_LAND:
			a = 'L';
			break;
		default:
			return GTC_RC_INVALIDARGUMENT;
	}
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":A%c#", a);
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_get_longitude (GtcScope *gs, double *longitude)
{
	int rc;	
	int deg;
	int am;
	int as;
	int haveac;
	double l;
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Gg#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 3;
	*d = 0;
	deg = atoi (p);
	p += 4;
	d += 3;
	haveac = 0;
	if (*d == ':') /* a : was found (arcseconds follow) */
		haveac = 1;
	*d = 0;
	am = atoi (p);
	if (haveac == 1) {
		p += 3;
		d += 3;
		*d = 0;
		as = atoi (p);
	} else {
		as = 0;
	}
	rc = degrees_dms_to_decimal (&l, deg, am, as);
	*longitude = l;
	return 0;
}


int meade_scope_set_longitude (GtcScope *gs, double longitude)
{
	int rc;	
	int deg;
	int am;
	int as;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = degrees_decimal_to_dms (longitude, &deg, &am, &as);
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Sg%03d*%02d#",
			deg, am);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad value */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_get_latitude (GtcScope *gs, double *latitude)
{
	int rc;	
	int deg;
	int am;
	int as;
	int haveac;
	double l;
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Gt#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 3;
	*d = 0;
	deg = atoi (p);
	p += 4;
	d += 3;
	haveac = 0;
	if (*d == ':') /* a : was found (arcseconds follow) */
		haveac = 1;
	*d = 0;
	am = atoi (p);
	if (haveac == 1) {
		p += 3;
		d += 3;
		*d = 0;
		as = atoi (p);
	} else {
		as = 0;
	}
	rc = degrees_dms_to_decimal (&l, deg, am, as);
	*latitude = l;
	return 0;
}


int meade_scope_set_latitude (GtcScope *gs, double latitude)
{
	int rc;	
	int deg;
	int am;
	int as;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = degrees_decimal_to_dms (latitude, &deg, &am, &as);
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":St%+02d*%02d#",
			deg, am);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad value */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_get_current_site (GtcScope *gs, int *site)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_set_current_site (GtcScope *gs, int site)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (site < 0 || site > 3)
		return GTC_RC_INVALIDARGUMENT;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":W%1d#",
			site);
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_east_start (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Me#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_east_stop (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Qe#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_west_start (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Mw#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_west_stop (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Qw#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_north_start (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Mn#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_north_stop (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Qn#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_south_start (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Ms#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_south_stop (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Qs#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_slew_stop (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Q#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_get_ra_slewrate (GtcScope *gs, double *degsec)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_set_ra_slewrate (GtcScope *gs, double degsec)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":RA%02.1f#",
			degsec);
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_get_dec_slewrate (GtcScope *gs, double *degsec)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_set_dec_slewrate (GtcScope *gs, double degsec)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":RE%02.1f#",
			degsec);
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_get_trackrate (GtcScope *gs, double *hertz)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GT#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	*hertz = atof (ms->buffer);
	return 0;
}


int meade_scope_set_trackrate (GtcScope *gs, double hertz)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":ST%02.1f#",
			hertz);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad value */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_get_guiderate (GtcScope *gs, double *arcsecsec)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_set_guiderate (GtcScope *gs, double arcsecsec)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Rg%02.1f#",
			arcsecsec);
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}



int meade_scope_focus_in_start (GtcScope *gs)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":F+#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_focus_out_start (GtcScope *gs)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":F-#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_focus_stop (GtcScope *gs)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":FQ#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_focus_get_speed (GtcScope *gs, double *speed)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_focus_set_speed (GtcScope *gs, double speed)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (speed < 1.0 || speed > 4.0)
		return GTC_RC_INVALIDARGUMENT;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":F%1d#",
			(int)floor(speed));
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_reticle_get_brightness (GtcScope *gs, double *brightness)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_reticle_set_brightness (GtcScope *gs, double brightness)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_reticle_get_flashrate (GtcScope *gs, double *rate)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_reticle_set_flashrate (GtcScope *gs, double rate)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (rate < 0.0 || rate > 9.0)
		return GTC_RC_INVALIDARGUMENT;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":B%1d#",
			(int)floor(rate));
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_reticle_get_dutyrate (GtcScope *gs, double *rate)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_reticle_set_dutyrate (GtcScope *gs, double rate)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (rate < 0.0 || rate > 15.0)
		return GTC_RC_INVALIDARGUMENT;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":BD%1d#",
			(int)floor(rate));
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_derotator_enable (GtcScope *gs, int enable)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (enable != 0)
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":r+#");
	else
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":r-#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_get_current_ra (GtcScope *gs, GtcRa *ra)
{
	int rc;	
	int havesec;
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GR#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 2;
	*d = 0;
	ra->hour = atoi (p);
	p += 3;
	d += 3;
	havesec = 0;
	if (*d == ':')
		havesec = 1;
	*d = 0;
	ra->minute = atoi (p);
	p += 3;
	if (havesec == 1)
		d += 3;
	else
		d += 2;
	*d = 0;
	if (havesec == 1)
		ra->second = atoi (p);
	else
		ra->second = atoi (p) * 6; /* tenths of an arcminute */
	return 0;
}


int meade_scope_get_current_dec (GtcScope *gs, GtcDec *dec)
{
	int rc;	
	int havesec;
	int deg;
	int minute;
	int second;
	double degrees;
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":GD#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 3;
	*d = 0;
	deg = atoi (p);
	p += 4;
	d += 3;
	havesec = 0;
	if (*d == ':')
		havesec = 1;
	*d = 0;
	minute = atoi (p);
	if (havesec == 1) {
		p += 3;
		d += 3;
		*d = 0;
		second = atoi (p);
	} else {
		second = 0;
	}
	rc = degrees_dms_to_decimal (&degrees,
				deg,
				minute,
				second);
	dec->deg = degrees;
	return 0;
}


int meade_scope_get_target_ra (GtcScope *gs, GtcRa *ra)
{
	int rc;	
	int havesec;
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Gr#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 2;
	*d = 0;
	ra->hour = atoi (p);
	p += 3;
	d += 3;
	havesec = 0;
	if (*d == ':')
		havesec = 1;
	*d = 0;
	ra->minute = atoi (p);
	p += 3;
	if (havesec == 1)
		d += 3;
	else
		d += 2;
	*d = 0;
	if (havesec == 1)
		ra->second = atoi (p);
	else
		ra->second = atoi (p) * 6; /* tenths of an arcminute */
	return 0;
}


int meade_scope_set_target_ra (GtcScope *gs, GtcRa *ra)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (ra->hour < 0 || ra->hour > 23 ||
	    ra->minute < 0 || ra->minute > 59 ||
	    ra->second < 0 || ra->second > 59)
		return GTC_RC_INVALIDARGUMENT;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Sr%02d:%02d:%02d#",
			ra->hour, ra->minute, ra->second);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad value */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_get_target_dec (GtcScope *gs, GtcDec *dec)
{
	int rc;	
	int havesec;
	int deg;
	int minute;
	int second;
	double degrees;
	char *p = NULL;
	char *d = NULL;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;

	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Gd#");
	rc = command_execute (ms, 0);
	if (rc != 0)
		return rc;
	p = ms->buffer;
	d = p;
	d += 3;
	*d = 0;
	deg = atoi (p);
	p += 4;
	d += 3;
	havesec = 0;
	if (*d == ':')
		havesec = 1;
	*d = 0;
	minute = atoi (p);
	if (havesec == 1) {
		p += 3;
		d += 3;
		*d = 0;
		second = atoi (p);
	} else {
		second = 0;
	}
	rc = degrees_dms_to_decimal (&degrees,
				deg,
				minute,
				second);
	dec->deg = degrees;
	return 0;
}


int meade_scope_set_target_dec (GtcScope *gs, GtcDec *dec)
{
	int rc;	
	int deg;
	int minute;
	int second;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (dec->deg < -90.0 || dec->deg > 90.0)
		return GTC_RC_INVALIDARGUMENT;
	rc = degrees_decimal_to_dms (dec->deg, &deg, &minute, &second);
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":Sd%+02d:%02d:%02d#",
			deg, minute, second);
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '0') /* bad value */
		return GTC_RC_INVALIDARGUMENT;
	return 0;
}


int meade_scope_goto_target (GtcScope *gs)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":MS#");
	rc = command_execute (ms, CMDFLAG_NOENDHASH);
	if (rc != 0)
		return rc;
	if (ms->buffer[0] == '1') /* Target Below Horizon */
		return GTC_RC_BELOWHORIZON;
	if (ms->buffer[0] == '2') /* Target Below Higher */
		return GTC_RC_HARDLIMIT;
	return 0;
}


int meade_scope_object_set_target (GtcScope *gs, int catalog, long object)
{
	int rc;
	char c;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (object < 0 || object > 9999)
		return GTC_RC_INVALIDARGUMENT;
	switch (catalog) {
		case GTC_CATALOG_DEEPSKY:
			c = 'C';
			break;
		case GTC_CATALOG_MESSIER:
			c = 'M';
			break;
		case GTC_CATALOG_STAR:
			c = 'S';
			break;
		default:
			return GTC_RC_INVALIDARGUMENT;
	}
	rc = snprintf (ms->buffer, BUFFER_SIZE, ":L%c%04lu#", c, object);
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_pec_ra_enable (GtcScope *gs, int enable)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (enable != 0)
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":QZ+#");
	else
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":QZ-#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_pec_dec_enable (GtcScope *gs, int enable)
{
	int rc;	
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (enable != 0)
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":QA+#");
	else
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":QA#");
	rc = command_execute (ms, CMDFLAG_NORESPONSE);
	if (rc != 0)
		return rc;
	return 0;
}


int meade_scope_option_list (GtcScope *gs, GtcOption **optionlist)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
			*optionlist = meade_options_lx200gps;
			break;
		default:
			*optionlist = NULL;
	}
	return 0;
}


int meade_scope_option_get_int (GtcScope *gs, char *name, int *value)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_option_set_int (GtcScope *gs, char *name, int value)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_option_get_double (GtcScope *gs, char *name, double *value)
{
	int rc;
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	if (ms->model == MEADE_MODEL_LX200GPS && (!strcmp (name, "OTATemp(C)"))) {
		rc = snprintf (ms->buffer, BUFFER_SIZE, ":fT#");
		rc = command_execute (ms, 0);
		if (rc != 0)
			return rc;
		*value = atof (ms->buffer);
		return 0;
	}
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_option_set_double (GtcScope *gs, char *name, double value)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_option_get_string (GtcScope *gs, char *name, char **value)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_option_set_string (GtcScope *gs, char *name, char *value)
{
	return GTC_RC_UNIMPLEMENTED;
}


int meade_scope_control_get_slewrates (GtcScope *gs, GtcControl **gcp)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
		case MEADE_MODEL_AUTOSTAR:
			*gcp = &meade_slewrates;
			break;
		default:
			*gcp = NULL;
	}
	return 0;
}


int meade_scope_control_get_trackrates (GtcScope *gs, GtcControl **gcp)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
		case MEADE_MODEL_AUTOSTAR:
		case MEADE_MODEL_LX200:
		case MEADE_MODEL_LX16:
			*gcp = &meade_trackrates;
			break;
		default:
			*gcp = NULL;
	}
	return 0;
}


int meade_scope_control_get_guiderates (GtcScope *gs, GtcControl **gcp)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
			*gcp = &meade_guiderates;
			break;
		default:
			*gcp = NULL;
	}
	return 0;
}


int meade_scope_control_get_focus_speeds (GtcScope *gs, GtcControl **gcp)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
		case MEADE_MODEL_AUTOSTAR:
			*gcp = &meade_focus_speeds;
			break;
		default:
			*gcp = NULL;
	}
	return 0;
}


int meade_scope_control_get_reticle_flashrates (GtcScope *gs, GtcControl **gcp)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
		case MEADE_MODEL_AUTOSTAR:
			*gcp = &meade_reticle_flashrates_lx200gps;
			break;
		case MEADE_MODEL_LX200:
			*gcp = &meade_reticle_flashrates_lx200;
			break;
		default:
			*gcp = NULL;
	}
	return 0;
}


int meade_scope_control_get_reticle_dutyrates (GtcScope *gs, GtcControl **gcp)
{
	MeadeScope *ms = NULL;
	ms = (MeadeScope *)gs->data;
	switch (ms->model) {
		case MEADE_MODEL_LX200GPS:
		case MEADE_MODEL_AUTOSTAR:
			*gcp = &meade_reticle_dutyrates;
			break;
		default:
			*gcp = NULL;
	}
	return 0;
}


